// File RTCG.cs 
// RTCG of methods that can be called efficiently via interface calls
// sestoft@dina.kvl.dk * 2005-10-16

// Compile with 
//   csc /t:library RTCG.cs
// Use as
//   csc /r:RTCG.dll RTCG2G.cs
//   csc /r:RTCG.dll AesGenRTCG.cs

// An Rtcg.ILFiller is a delegate that takes as argument an
// ILGenerator; the purpose of an ILFiller is to populate the
// body of a method with bytecode instructions.  In other words, 
// type ILFiller = ILGenerator -> unit.

// Method MethodObject is called as 
//   MethodObject(ilFiller, iface, rTy, argTy1, ..., argTyN) 
// and returns an object of a new class.  The new class implements 
// interface iface and contains a virtual method with return type rTy 
// and parameters of type argTy1, ..., argTyN where N>=0.  This is an 
// auxiliary method used in e.g. the following method:

// IFun<A1,R> RtcgFun<A1,R>(ILFiller ilFiller) returns an object of a
// class that implements interface IFun<A1,R>.  It therefore has an
// instance method Invoke with signature 
//    R Invoke(A1)
// and the body of that method was generated by a call to ilFiller.

// IAct<A1> RtcgAct<A1>(ILFiller ilFiller) returns an object of a
// class that implements interface IAct<A1>.  It therefore has an
// instance method Invoke with signature 
//    void Invoke(A1)
// and the body of that method was generated by a call to ilFiller.

using System;
using System.Reflection;
using System.Reflection.Emit;

namespace Rtcg {
  // Interfaces describing non-void methods generated by RTCG
  public interface IFun<R> { R Invoke(); }
  public interface IFun<A1,R> { R Invoke(A1 x); }
  public interface IFun<A1,A2,R> { R Invoke(A1 x1, A2 x2); }
  public interface IFun<A1,A2,A3,R> { R Invoke(A1 x1, A2 x2, A3 x3); }
  
  // Interfaces describing void methods generated by RTCG
  public interface IAct { void Invoke(); }
  public interface IAct<A1> { void Invoke(A1 x); }
  public interface IAct<A1,A2> { void Invoke(A1 x1, A2 x2); }
  public interface IAct<A1,A2,A3> { void Invoke(A1 x1, A2 x2, A3 x3); }
  
  // The type of delegates for building the body of a generated method
  public delegate void ILFiller(ILGenerator ilg);

  // Static class for manufacturing classes that contain runtime methods
  
  public static class RTCG {
    private static ModuleBuilder moduleBuilder = null;
    private static int classNumber = 0;
    
    // Create and cache a single AssemblyBuilder and ModuleBuilder
    
    private static ModuleBuilder GetModuleBuilder { 
      get { 
	if (moduleBuilder == null) {
	  AssemblyName assemblyName = new AssemblyName();
	  assemblyName.Name = "myassembly";
	  // Build: run-only assembly
	  AssemblyBuilder assemblyBuilder = 
	    AppDomain.CurrentDomain.DefineDynamicAssembly
	    (assemblyName, AssemblyBuilderAccess.Run);
	  // Build: module mymodule
	  moduleBuilder = assemblyBuilder.DefineDynamicModule("mymodule");
	}
	return moduleBuilder;
      }
    }

    // Entry methods to generate non-void methods using RTCG

    public static IFun<R> RtcgFun<R>(ILFiller ilFiller) {
      return (IFun<R>)
	MethodObject(ilFiller, typeof(IFun<R>), typeof(R));
    }
    
    public static IFun<A1,R> RtcgFun<A1,R>(ILFiller ilFiller) {
      return (IFun<A1,R>)
	MethodObject(ilFiller, typeof(IFun<A1,R>), typeof(R), typeof(A1));
    }
    
    public static IFun<A1,A2,R> RtcgFun<A1,A2,R>(ILFiller ilFiller) {
      return (IFun<A1,A2,R>)
	MethodObject(ilFiller, typeof(IFun<A1,A2,R>), 
		     typeof(R), typeof(A1), typeof(A2));
    }

    // Entry methods to generate void methods using RTCG

    public static IAct RtcgAct(ILFiller ilFiller) {
      return (IAct)
	MethodObject(ilFiller, typeof(IAct), typeof(void));
    }

    public static IAct<A1> RtcgAct<A1>(ILFiller ilFiller) {
      return (IAct<A1>)
	MethodObject(ilFiller, typeof(IAct<A1>), typeof(void), typeof(A1));
    }
    
    public static IAct<A1,A2> RtcgAct<A1,A2>(ILFiller ilFiller) {
      return (IAct<A1,A2>)
	MethodObject(ilFiller, typeof(IAct<A1,A2>), 
		     typeof(void), typeof(A1), typeof(A2));
    }

    // General method to generate, using RTCG, a method in a new class
    // so that the method is efficiently callable through an interface
    
    private static Object MethodObject(ILFiller ilFiller, Type iface, 
		 		       Type returnTy, params Type[] argTy) {
      // Build: public class MyClassN : iface { ... }
      TypeBuilder typeBuilder = 
	GetModuleBuilder.DefineType("MyClass" + (classNumber++),
				    TypeAttributes.Class | TypeAttributes.Public,
				    typeof(Object));
      typeBuilder.AddInterfaceImplementation(iface);
      // Build constructor: public MyClassN() : base() { }
      ConstructorBuilder constructorBuilder = 
	typeBuilder.DefineConstructor(MethodAttributes.Public,
				      CallingConventions.HasThis,
				      new Type[] { });
      ILGenerator ilg = constructorBuilder.GetILGenerator();
      ilg.Emit(OpCodes.Ldarg_0);      // push this
      ilg.Emit(OpCodes.Call, typeof(Object).GetConstructor(new Type[] {}));
      ilg.Emit(OpCodes.Ret);
      // Build method: public virtual returnTy Invoke(argTy) { ... }
      MethodBuilder methodBuilder =
	typeBuilder.DefineMethod("Invoke", 
				 MethodAttributes.Virtual 
				 | MethodAttributes.Public,
				 returnTy,
				 argTy);
      // Obtain an IL generator and build the method body
      ilFiller(methodBuilder.GetILGenerator());
      // Create the class type and create an instance of it
      Type ty = typeBuilder.CreateType();
      return ty.GetConstructor(new Type[] {}).Invoke(new Object[] { });
    }
  }
}
